/*
	Copyright (c) 2004 Pablo Bleyer Kocik.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	1. Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

	3. The name of the author may not be used to endorse or promote products
	derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
	WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
	EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
	IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
*/

import java.util.Vector;
import java.util.TreeSet;
import java.util.ListIterator;
import java.util.Iterator;

public class Assembler {

	public static final int
		program_size = 1<<10;

	public int[] program;
	public String[] listing;
	public Environment environment;

	Assembler() {
		environment = new Environment();
		program = new int[program_size];
		listing = new String[program_size];
		for (int i=0; i<program.length; ++i) program[i] = 0;
	}

	/** Parse a vector of commands for symbols. */
	void parse(Vector l) {
		int pc = 0; // program counter
		ListIterator i = l.listIterator();

		while (i.hasNext()) {
			Command c = (Command)i.next();
			if (c.empty()) continue;

			// add label
			if (c.label != null)
				environment.symbols.put(c.label, new Symbol(Symbol.LABEL, pc));

			if (c.name != null) {
				// add constants
				if (c.name.equals("constant")) {
					if (c.list == null
						|| c.list.size() != 2
						|| !(c.list.get(0) instanceof String)
						|| !(c.list.get(1) instanceof Integer)
						) throw new IllegalArgumentException();
					environment.symbols.put(
						c.list.get(0), new Symbol(Symbol.CONSTANT, c.list.get(1))
						);
					// overwrite label if it exists
					if (c.label != null)
						environment.symbols.put(
							c.label, new Symbol(Symbol.LABEL, c.list.get(1))
							);
				}
				else if (c.name.equals("address")) {
					if (c.list == null
						|| c.list.size() != 1
						|| !(c.list.get(0) instanceof Integer)
						) throw new IllegalArgumentException();
					pc = ((Integer)c.list.get(0)).intValue();
					// overwrite label if it exists
					if (c.label != null)
						environment.symbols.put(
							c.label, new Symbol(Symbol.LABEL, pc)
							);
				}
				else if (c.name.equals("register")) {
					if (c.list == null
						|| c.list.size() != 2
						|| !(c.list.get(0) instanceof String)
						|| !(c.list.get(1) instanceof Integer)
						) throw new IllegalArgumentException("Bad register declaration");
					environment.symbols.put(
						c.list.get(0), new Symbol(Symbol.REGISTER, c.list.get(1))
						);
					// overwrite label
					if (c.label != null)
						environment.symbols.put(
							c.label, new Symbol(Symbol.LABEL, c.list.get(1))
							);
				}
				else ++pc;
			}
		}
	}

	/** Assemble a vector of commands. */
	void assemble(Vector l) {
		int pc = 0; // program counter
		ListIterator i = l.listIterator();

		while (i.hasNext()) {
			Command c = (Command)i.next();
			if (c.empty()) continue;

			if (c.name != null) {
				if (c.name.equals("address")) {
					pc = ((Integer)c.list.get(0)).intValue();
				}
				else if (!c.name.equals("constant")
					&& !c.name.equals("register")
					) {
					Instruction t = new Instruction(c.name, c.list, environment);
					program[pc] = t.toCode();
					listing[pc] = c.toString();
					++pc;
				}
			}
		}

	}

	public String toString() {
		StringBuffer b = new StringBuffer();
		Iterator i = (new TreeSet(environment.symbols.keySet())).iterator();

		b.append("// <Symbol Table>\n");
		while (i.hasNext()) {
			String k = (String)i.next();
			b.append ("// " + k + " = ");
			Symbol s = (Symbol)environment.symbols.get(k);
			if (s != null) b.append(s.toString());
			b.append("\n");
		}

		b.append("\n");
		b.append("// <Program Code>\n");

		for (int j=0; j<program_size; ++j) {
			String h = Integer.toHexString(program[j]);
			int l = 5-h.length();
			for (int m=0; m<l; ++m) h = "0" + h;
			b.append(h);
			// b.append(String.format("%1$x.5", program[j]));
			if (listing[j] != null) b.append(" // " + listing[j]);
			b.append("\n");
		}

		return b.toString();
	}

}
